@@ -1,16 +1,19 @@
 #' Sequence weighting.
 #'
-#' Weight sequences based on a tree.
+#' Weighting schemes for DNA and amino acid sequences.
 #'
-#' @param x an object of class \code{"dendrogram"}, or a list or matrix of
-#'   sequences (possibly a "DNAbin" or "AAbin" object) from which to derive a
-#'   dendrogram using the \code{\link[kmer]{cluster}} function in the
-#'   \code{\link[kmer]{kmer}} package.
+#' @param x a list or matrix of sequences
+#'   (usually a "DNAbin" or "AAbin" object).
+#'   Alternatively x can be an object of class \code{"dendrogram"}
+#'   for tree-base weighting.
 #' @param method a character string indicating the weighting method to be used.
-#'   Currently only that of Gerstein et al (1994) is supported
+#'   Currently the only methods available are a modified version of the
+#'   maximum entropy weighting scheme proposed by
+#'   Henikoff and Henikoff (1994) (\code{method = "Henikoff"})
+#'   and the tree-based weighting scheme of Gerstein et al (1994)
 #'   (\code{method = "Gerstein"}).
-#' @param k integer representing the k-mer size to be used in tree-based
-#'   sequence weighting. Defaults to 5. Note that higher
+#' @param k integer representing the k-mer size to be used.
+#'   Defaults to 5. Note that higher
 #'   values of k may be slow to compute and use excessive memory due to
 #'   the large numbers of calculations required.
 #' @param residues either NULL (default; emitted residues are automatically
@@ -29,14 +32,25 @@
 #' @return a named vector of weights, the sum of which is equal to
 #'    the total number of sequences (average weight = 1).
 #' @details
-#'   This is a generic function that uses the agglomerative method of
-#'   Gerstein et al (1994) to weight sequences based on their relatedness
-#'   as derived from a phylogenetic tree. Methods are available for
+#'   This is a generic function.
+#'   If \code{method = "Henikoff"} the sequences are weighted
+#'   using a modified version of the maximum entropy method proposed by
+#'   Henikoff and Henikoff (1994). In this case the
+#'   maximum entropy weights are calculated from a k-mer presence absence
+#'   matrix instead of an alignment as originally described by
+#'   Henikoff and Henikoff (1994).
+#'   If \code{method = "Gerstein"} the agglomerative method of
+#'   Gerstein et al (1994) is used to weight sequences based
+#'   on their relatedness as derived from a phylogenetic tree.
+#'   In this case a dendrogram is first derived using the
+#'   \code{\link[kmer]{cluster}} function in the
+#'   \code{\link[kmer]{kmer}} package.
+#'   Methods are available for
 #'   \code{"dendrogram"} objects, \code{"DNAbin"} and \code{"AAbin"}
 #'   sequence objects (as lists or matrices) and sequences in standard
-#'   ASCII character format provided either as lists or matrices.
+#'   character format provided either as lists or matrices.
 #'
-#'   For further details on sequence weighting see Durbin et al
+#'   For further details on sequence weighting schemes see Durbin et al
 #'   (1998) chapter 5.8.
 #'
 #' @author Shaun Wilkinson
@@ -48,7 +62,9 @@
 #'   Gerstein M, Sonnhammer ELL, Chothia C (1994) Volume changes in protein evolution.
 #'   \emph{Journal of Molecular Biology}, \strong{236}, 1067-1078.
 #'
-#' @seealso \code{\link[kmer]{cluster}}
+#'   Henikoff S, Henikoff JG (1994) Position-based sequence weights.
+#'   \emph{Journal of Molecular Biology}, \strong{243}, 574-578.
+#'
 #' @examples
 #'   ## weight the sequences in the woodmouse dataset from the ape package
 #'   library(ape)
@@ -63,7 +79,7 @@ weight <- function(x, ...){
 ################################################################################
 #' @rdname weight
 ################################################################################
-weight.DNAbin <- function(x, method = "Gerstein", k = 5, ...){
+weight.DNAbin <- function(x, method = "Henikoff", k = 5, ...){
   if(is.list(x)){
     weight.list(x, method = method, k = k)
   }else{
@@ -74,7 +90,7 @@ weight.DNAbin <- function(x, method = "Gerstein", k = 5, ...){
 ################################################################################
 #' @rdname weight
 ################################################################################
-weight.AAbin <- function(x, method = "Gerstein", k = 5, ...){
+weight.AAbin <- function(x, method = "Henikoff", k = 5, ...){
   if(is.list(x)){
     weight.list(x, method = method, k = k)
   }else{
@@ -85,7 +101,7 @@ weight.AAbin <- function(x, method = "Gerstein", k = 5, ...){
 ################################################################################
 #' @rdname weight
 ################################################################################
-weight.list <- function(x, method = "Gerstein", k = 5, residues = NULL,
+weight.list <- function(x, method = "Henikoff", k = 5, residues = NULL,
                         gap = "-", ...){
   nsq <- length(x)
   DNA <- .isDNA(x)
@@ -94,20 +110,40 @@ weight.list <- function(x, method = "Gerstein", k = 5, residues = NULL,
   residues <- .alphadetect(x, residues = residues, gap = gap)
   gap <- if(AA) as.raw(45) else if(DNA) as.raw(4) else gap
   for(i in 1:nsq) x[[i]] <- x[[i]][x[[i]] != gap]
-  # cache names for later
-  tmpnames <- names(x)
-  names(x) <- paste0("S", 1:nsq)
   if(nsq > 2){
-    guidetree <- kmer::cluster(x, k = k, residues = residues, gap = gap)
-    res <- weight.dendrogram(guidetree, method = "Gerstein")[names(x)]
+    if(identical(method, "Henikoff")){
+      hashes <- .digest(x)
+      pointers <- .point(hashes)
+      catchnames <- names(x)
+      x <- x[!duplicated(pointers)]
+      nsq <- length(x)
+      suppressMessages(kmers <- round(kmer::kcount(x, k = k, residues = residues, gap = gap)))
+      kmers <- kmers > 0
+      ksums <- apply(kmers, 2, sum)
+      uwfs <- ksums/nsq # unweighted freqs
+      ftweights <- rbind(1/(2*(1-uwfs)), 1/(2*uwfs))
+      f <- function(a, b) mean(b[rbind(!a, a)])
+      res <- apply(kmers, 1, f, ftweights)
+      rm(kmers)
+      spl <- split(seq_along(hashes), f = factor(pointers))
+      ## divide weight eveny among duplicates
+      res <- res/vapply(spl, length, 0L)
+      res <- res[pointers]
+      names(res) <- catchnames
+    }else if(identical(method, "Gerstein")){
+      tmpnames <- names(x)
+      names(x) <- paste0("S", 1:nsq)
+      guidetree <-suppressMessages(kmer::cluster(x, k = k, residues = residues, gap = gap))
+      res <- weight.dendrogram(guidetree, method = "Gerstein")[names(x)]
+      names(res) <- tmpnames
+    }else stop("Accepted methods are 'Gerstein' and 'Henikoff'\n")
   }else if(nsq == 2){
-    res <- c(1, 1)
+    res <- structure(c(1, 1), names = names(x))
   }else if(nsq == 1){
-    res <- 1
+    res <- structure(1, names = names(x))
   }else{
     res <- numeric(0)
   }
-  names(res) <- tmpnames
   return(res)
 }
 ################################################################################
@@ -174,7 +210,7 @@ weight.dendrogram <- function(x, method = "Gerstein", ...){
 ################################################################################
 #' @rdname weight
 ################################################################################
-weight.default <- function(x, method = "Gerstein", k = 5, residues = NULL,
+weight.default <- function(x, method = "Henikoff", k = 5, residues = NULL,
                            gap = "-", ...){
   x <- unalign(x, gap = gap)
   weight.list(x, method = method, k = k, residues = residues, gap = gap)
